import json
import os
import tempfile
import time

import click

from jycm.helper import dump_html_output, make_ignore_order_func
from jycm.jycm import YouchamaJsonDiffer
from jycm.operator import get_operator


def diff_two_json_with_rules(left, right, rules, debug=False, fast_mode=False):
    operators = []
    ignore_order_path_regex_list = []
    for rule in rules:
        operation = rule["operation"]
        path_regex = rule["value"]
        parameter = rule.get("parameter", {})

        if parameter is None:
            parameter = {}

        if operation == "operator:list:ignoreOrder":
            ignore_order_path_regex_list.append(
                path_regex
            )
            continue

        operator_class = get_operator(operation)
        operators.append(operator_class(path_regex, **parameter))
    ycm = YouchamaJsonDiffer(
        left, right,
        ignore_order_func=make_ignore_order_func(ignore_order_path_regex_list),
        custom_operators=operators,
        debug=debug,
        fast_mode=fast_mode
    )
    same = ycm.diff()

    return same, ycm.to_dict()


def load_json(raw):
    try:
        return json.loads(raw)
    except Exception as e:
        try:
            return json.loads(raw.replace("\'", "\""))
        except Exception as e:
            raise ValueError(f"Not a valid json: `{raw}`")


def run(left, right, rules, output, show):
    left = load_json(left)
    right = load_json(right)
    rules = load_json(rules)

    same, result = diff_two_json_with_rules(left, right, rules)

    if output is not None:
        index_url = dump_html_output(left, right, result, output)
        if show:
            try:
                import webbrowser
                from sys import platform
                if platform == "linux" or platform == "linux2":
                    webbrowser.open(f"file://{index_url}")
                elif platform == "darwin":
                    webbrowser.open(f"file://{index_url}")
                elif platform == "win32":
                    webbrowser.open(f"{index_url}")
            except Exception as e:
                print("You have to install webbrowser to open the html.\nRun pip install webbrowser")

    print("No diff" if same else "Has diff!")


def get_json_input(key):
    while True:
        val = input(f"{key} >>>")
        try:
            json.loads(val)
            return val
        except json.decoder.JSONDecodeError:
            print(f"Not a valid json: {val}")
            continue
        except Exception as e:
            raise(e)


def interactive_main(output):
    if output is None:
        output = tempfile.mkdtemp()

    while True:
        try:
            left = get_json_input("Left Json")
            right = get_json_input("Right Json")
            rules = get_json_input("Rules")
            run(left, right, rules, os.path.join(output, f"jycm-{int(time.time())}"), True)
        except Exception as e:
            print("Exception", e)
            pass


@click.command()
@click.option('--interactive', is_flag=True, show_default=True, help='Enter interactive mode')
@click.option('--left', default='{}', help='Left Json')
@click.option('--right', default='{}', help='Right Json')
@click.option('--rules', default='[]', help='Rules')
@click.option('--output', default=None, help='The folder where the results will be dumped.')
@click.option('--show', is_flag=True, show_default=True, help='Whether or not open the browser to visualize result.')
def main(interactive, left, right, rules, output, show):
    print(interactive, left, right, rules, output, show)
    if interactive:
        return interactive_main(output)

    if output is None:
        output = os.path.join(tempfile.mkdtemp(), f"jycm-{int(time.time())}")

    return run(left, right, rules, output, show)


if __name__ == '__main__':
    main()



# coding: utf-8

"""jycm codes. .
"""

from __future__ import absolute_import, division, print_function, unicode_literals

from .__version__ import __version__  # NOQA

__author__ = 'eggachecat <sunao_0626@hotmail.com>'



import logging
import re
from typing import TYPE_CHECKING, Tuple, Type

from jycm.common import PLACE_HOLDER_NON_EXIST

if TYPE_CHECKING:
    from jycm.jycm import TreeLevel, YouchamaJsonDiffer


class BaseOperator:
    __operator_name__ = "__base__"

    def __init__(self, path_regex: str):
        self.path_regex = path_regex
        self.regex = re.compile(f"{self.path_regex}")

    def match(self, level: 'TreeLevel') -> bool:
        matched = self.regex.search(level.get_path()) is not None
        return matched

    def diff(self, level: 'TreeLevel', instance, drill: bool) -> Tuple[bool, float]:
        raise NotImplementedError


OPERATOR_DICT = {}


def register_operator(operator_class: Type[BaseOperator]):
    if isinstance(operator_class.__operator_name__, str):
        operator_name_list = [operator_class.__operator_name__]
    elif isinstance(operator_class.__operator_name__, list):
        operator_name_list = operator_class.__operator_name__
    else:
        raise TypeError("bad type for ", operator_class.__operator_name__)

    for name in operator_name_list:
        if name in OPERATOR_DICT:
            # just a warning allow overriding
            logging.warning(f"duplicate operator name: {operator_class.__operator_name__}")

        OPERATOR_DICT[name] = operator_class

    return operator_class


def get_operator(name: str):
    if name not in OPERATOR_DICT:
        logging.warning(f"unknown operation=[{name}]")

    return OPERATOR_DICT[name]


@register_operator
class ListItemFieldMatchOperator(BaseOperator):
    __operator_name__ = "operator:list:matchWithField"
    __event__ = "operator:list:matchWithField"

    def __init__(self, path_regex, field):
        super().__init__(path_regex=path_regex)
        self.field = field

    def diff(self, level: 'TreeLevel', instance: 'YouchamaJsonDiffer', drill: bool) -> Tuple[bool, float]:
        if drill:
            # 演习的比较
            if level.left[self.field] == level.right[self.field]:
                return True, 1
        else:
            instance.report(self.__event__, level, {"field": self.field, "path_regex": self.path_regex})

        return False, -1


@register_operator
class ExpectChangeOperator(BaseOperator):
    __operator_name__ = "operator:expectChange"
    __event__ = "operator:expectChange"

    def diff(self, level: 'TreeLevel', instance: 'YouchamaJsonDiffer', drill: bool) -> Tuple[bool, float]:
        if level.left == level.right:
            if not drill:
                instance.report(self.__event__, level, {"pass": False, "path_regex": self.path_regex})
            return True, 0

        if not drill:
            instance.report(self.__event__, level, {"pass": True, "path_regex": self.path_regex})

        return True, 1


@register_operator
class ExpectExistOperator(BaseOperator):
    __operator_name__ = "operator:expectExist"
    __event__ = "operator:expectExist"

    def diff(self, level: 'TreeLevel', instance: 'YouchamaJsonDiffer', drill: bool) -> Tuple[bool, float]:

        info = {
            "pass": True,
            "path_regex": self.path_regex
        }

        if level.left == PLACE_HOLDER_NON_EXIST:
            info["pass"] = False
            info["left_non_exist"] = True

        if level.right == PLACE_HOLDER_NON_EXIST:
            info["pass"] = False
            info["right_non_exist"] = True

        # 只要到这里就ok
        if not drill:
            instance.report(self.__event__, level, info)

        if info["pass"]:
            return True, 1

        return True, 0


@register_operator
class FloatInRangeOperator(BaseOperator):
    __operator_name__ = "operator:floatInRange"
    __event__ = "operator:floatInRange"

    def __init__(self, path_regex, interval_start, interval_end):
        super().__init__(path_regex=path_regex)
        self.interval_start = interval_start
        self.interval_end = interval_end

    def diff(self, level: 'TreeLevel', instance: 'YouchamaJsonDiffer', drill: bool) -> Tuple[bool, float]:
        info = {
            "interval_start": self.interval_start,
            "interval_end": self.interval_end,
            "path_regex": self.path_regex,
            "pass": True
        }

        invalid = False
        if not (self.interval_start < level.left <= self.interval_end):
            info.update({"left_invalid": True, "pass": False})
            invalid = True

        if not (self.interval_start < level.right <= self.interval_end):
            info.update({"right_invalid": True, "pass": False})
            invalid = True

        if not drill:
            instance.report(self.__event__, level, info)

        if invalid:
            return True, 0

        # 只要到这里就ok
        return True, 1


@register_operator
class IgnoreOperator(BaseOperator):
    __operator_name__ = [
        "ignore", "diff_word", "diff_image", "diff_pdf"
    ]
    __event__ = "ignore"

    def __init__(self, path_regex: str, *args, **kwargs):
        super().__init__(path_regex)

    def diff(self, level: 'TreeLevel', instance: 'YouchamaJsonDiffer', drill: bool) -> Tuple[bool, float]:
        info = {
            "path_regex": self.path_regex,
            "pass": True
        }
        if not drill:
            instance.report(self.__event__, level, info)

        return True, 1



# coding: utf-8

import io
import os

from setuptools import find_packages, setup

# 包元信息
NAME = 'jycm'  # 实际包的名字
DESCRIPTION = 'A highly flexible json diff framework for python.'  # 项目描述
URL = 'https://github.com/eggachecat/jycm'  # 项目仓库 URL
EMAIL = 'sunao_0626@hotmail.com'  # 维护者邮箱地址
AUTHOR = 'eggachecat'  # 维护者姓名

here = os.path.abspath(os.path.dirname(__file__))


def get_requirements():
    with open(os.path.join(here, "requirements.txt"), "r") as reqs_file:
        reqs = reqs_file.readlines()
    return reqs


def get_requirements_dev():
    final_reqs = get_requirements()
    with open(os.path.join(here, "requirements-dev.txt"), "r") as reqs_file:
        reqs = reqs_file.readlines()
    for r in reqs:
        if "-r" not in r:
            final_reqs.append(r)

    return final_reqs


REQUIRES = get_requirements()
DEV_REQUIRES = get_requirements_dev()

try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except IOError:
    long_description = DESCRIPTION

about = {}
with io.open(os.path.join(here, NAME, '__version__.py')) as f:
    exec(f.read(), about)

setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='json diff jsondiff operator flexible',
    packages=find_packages(exclude=['docs', 'tests']),
    install_requires=REQUIRES,
    tests_require=[
        'pytest>=4.0.0,<5.0.0'
    ],
    python_requires='!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
    extras_require={
        ':python_version<"3.5"': [
            'typing>=3.6.4',
        ],
        'dev': DEV_REQUIRES,
    },
    package_data={
        # for PEP484 & PEP561
        NAME: ['py.typed', '*.pyi'],
    },
    include_package_data=True
)
